# -*- Mode: Python; tab-width: 4 -*-

import asynchat
import socket
import string
import time         # these three are for the rotating logger
import os           # |
import stat         # v

#
# three types of log:
# 1) file
#    with optional flushing.  Also, one that rotates the log.
# 2) socket
#    dump output directly to a socket connection. [how do we
#    keep it open?]
# 3) syslog
#    log to syslog via tcp.  this is a per-line protocol.
#

#
# The 'standard' interface to a logging object is simply
# log_object.log (message)
#

# a file-like object that captures output, and
# makes sure to flush it always...  this could
# be connected to:
#  o	stdio file
#  o	low-level file
#  o	socket channel
#  o	syslog output...

class file_logger:

        # pass this either a path or a file object.
    def __init__ (self, file, flush=1, mode='a'):
        if type(file) == type(''):
            if (file == '-'):
                import sys
                self.file = sys.stdout
            else:
                self.file = open (file, mode)
        else:
            self.file = file
        self.do_flush = flush
        
    def __repr__ (self):
        return '<file logger: %s>' % self.file
        
    def write (self, data):
        self.file.write (data)
        self.maybe_flush()
        
    def writeline (self, line):
        self.file.writeline (line)
        self.maybe_flush()
        
    def writelines (self, lines):
        self.file.writelines (lines)
        self.maybe_flush()
        
    def maybe_flush (self):
        if self.do_flush:
            self.file.flush()
            
    def flush (self):
        self.file.flush()
        
    def softspace (self, *args):
        pass
        
    def log (self, message, host = None):
        if message[-1] not in ('\r', '\n'):
            self.write (message + '\n')
        else:
            self.write (message)
            

class virtual_host_file_logger (file_logger):

    expirytime = 7200   # 2 hours
    no_host_name = 'no_host'

    def __init__ (self, dir, flush=1, mode='a', hosts=[], file_format=0):
        self.files = {}
        self.files_last_accessed = {}

        # remove any extension on the filename and add '_logs' eg 'Z2.log' --> 'Z2_logs'
        (temppath, tempfile) = os.path.split(dir)
        (tempfilename, tempfileext) = os.path.splitext(tempfile)
        if tempfilename == '':
            tempfilename = 'logs'
        else:
            tempfilename = '%s_logs' % tempfilename
        dir = os.path.join(temppath, tempfilename)

        # if dir doesn't exist, make it
        if not os.path.isdir(dir):
            os.makedirs(dir)

        # get allowed virtual hosts list
        self.get_virtual_hosts(hosts)

        # save settings
        self.basedir = dir
        self.mode = mode
        self.do_flush = flush
        self.file_format = file_format

        # set expirytime
        self.check_for_expired = time.time() + self.expirytime

    def __repr__ (self):
        return '<virtual host file logger>'

    def reopen(self):
        for host in self.files.keys():
            self.files[host].close()
            del self.files[host]
            get_file(host)

    def maybe_flush (self, host = ''):
        if self.do_flush:
            self.get_file(host).flush()

    def flush (self, host = ''):
        self.get_file(host).flush()

    def write (self, data, host = ''):
        self.get_file(host).write (data)
        self.maybe_flush(host)

    def writeline (self, line, host = ''):
        self.get_file(host).writeline (line)
        self.maybe_flush(host)

    def writelines (self, lines, host = ''):
        self.get_file(host).writelines (lines)
        self.maybe_flush(host)

    def log (self, message, host = ''):
        if message[-1] not in ('\r', '\n'):
            self.write (message + '\n', host)
        else:
            self.write (message, host)

    def get_filename (self, host, date=None):
        # if file_format is true then we are using the host as a directory
        if self.file_format:
            filename = os.path.join(host, date or 'current')
        else:
            filename = host
        return os.path.join(self.basedir, '%s.log' % (filename))
    
    # Choose the right open file or open it now
    def get_file (self, host = ''):
        host = self.clean_up_host(host)
        openhosts = self.files.keys()
        
        now = time.time()
        # Regardless whether the file is open or not, it has just been "accessed"
        self.files_last_accessed[host] = now
        
        # Close expired files (but only do this check every expirytime since it's
        #    too time consuming to be done every time we try to access a file!)
        if self.check_for_expired < now:
            expired = now - self.expirytime
            for checkhost in openhosts:
                if self.files_last_accessed[checkhost] < expired:
                    self.files[checkhost].close()
                    del self.files[checkhost]
                    del self.files_last_accessed[checkhost]
            self.check_for_expired = now + self.expirytime

        # Return the file (open if necessary)
        if not host in openhosts:
            filename = self.get_filename(host)
            dir = os.path.dirname(filename)
            # if dir doesn't exist, make it
            if not os.path.isdir(dir):
                os.makedirs(dir)
            file = open(filename, self.mode)
            self.files[host] = file
        else:
            file = self.files[host]
        return file

    def clean_up_host(self, host):
        # get rid of surrounding spaces
        host = host.strip()

        # strip off the 'www.'
        if host.startswith('www.'):
            host = host[4:]

        # rid ourselves of any port (this will meld logs from different ports
        #    & same host however - perhaps this check should be removed?)
        colon = host.rfind(':')
        if colon != -1:
            host = host[:colon]

        # safety check, replace all non allowed characters(is this pointlessly careful?)
        for char in '\\/*?"<>|:':
            host = host.replace(char, '_')  # perhaps _ isn't best in case this is the first character?
            
        if self.allowed_virtual_hosts:
            if not host in self.allowed_virtual_hosts:
                host = self.no_host_name

        if not host:
            host = self.no_host_name

        return host

    def get_virtual_hosts(self, virtual_hosts):        
        # clear the list so we don't get in to trouble with clean_up_host
        self.allowed_virtual_hosts = []
        # clean up the hosts list
        virtual_hosts = map(self.clean_up_host, virtual_hosts)
        # remove non-host name
        if self.no_host_name in virtual_hosts:
            virtual_hosts.remove(self.no_host_name)

        # remove duplicates
        set = []
        [set.append(item) for item in virtual_hosts if not item in set]
        virtual_hosts = set
        
        self.allowed_virtual_hosts = virtual_hosts


# like a file_logger, but it must be attached to a filename.
# When the log gets too full, or a certain time has passed,
# it backs up the log and starts a new one.  Note that backing
# up the log is done via "mv" because anything else (cp, gzip)
# would take time, during which medusa would do nothing else.
            
class rotating_file_logger (file_logger):

        # If freq is non-None we back up "daily", "weekly", or "monthly".
        # Else if maxsize is non-None we back up whenever the log gets
        # to big.  If both are None we never back up.
    def __init__ (self, file, freq=None, maxsize=None, flush=1, mode='a'):
        self.filename = file
        self.mode = mode
        self.file = open (file, mode)
        self.freq = freq
        self.maxsize = maxsize
        self.rotate_when = self.next_backup(self.freq)
        self.do_flush = flush
        
    def __repr__ (self):
        return '<rotating-file logger: %s>' % self.file
        
        # We back up at midnight every 1) day, 2) monday, or 3) 1st of month
    def next_backup (self, freq):
        (yr, mo, day, hr, min, sec, wd, jday, dst) = time.localtime(time.time())
        if freq == 'daily':
            return time.mktime((yr,mo,day+1, 0,0,0, 0,0,-1))
        elif freq == 'weekly':
            return time.mktime((yr,mo,day-wd+7, 0,0,0, 0,0,-1))  # wd(monday)==0
        elif freq == 'monthly':
            return time.mktime((yr,mo+1,1, 0,0,0, 0,0,-1))
        else:
            return None                  # not a date-based backup
            
    def maybe_flush (self):              # rotate first if necessary
        self.maybe_rotate()
        if self.do_flush:                # from file_logger()
            self.file.flush()
            
    def maybe_rotate (self):
        if self.freq and time.time() > self.rotate_when:
            self.rotate()
            self.rotate_when = self.next_backup(self.freq)
        elif self.maxsize:               # rotate when we get too big
            try:
                if os.stat(self.filename)[stat.ST_SIZE] > self.maxsize:
                    self.rotate()
            except os.error:             # file not found, probably
                self.rotate()            # will create a new file
                
    def rotate (self):
        (yr, mo, day, hr, min, sec, wd, jday, dst) = time.localtime(time.time())
        try:
            self.file.close()
            newname = '%s.ends%04d%02d%02d' % (self.filename, yr, mo, day)
            try:
                open(newname, "r").close()      # check if file exists
                newname = newname + "-%02d%02d%02d" % (hr, min, sec)
            except:                             # YEARMODY is unique
                pass
            os.rename(self.filename, newname)
            self.file = open(self.filename, self.mode)
        except:
            pass
            

class rotating_virtual_host_file_logger (virtual_host_file_logger, rotating_file_logger):

    # If freq==None, close file after timeout!
    def __init__ (self, dir, freq=None, flush=1, mode='a', hosts=[], file_format=0):
        virtual_host_file_logger.__init__ (self, dir, flush=1, mode='a', hosts=hosts, file_format=file_format)
        self.freq = freq
        self.rotate_when = self.next_backup(self.freq)
        self.check_for_expired = self.next_backup('daily')
        
    def __repr__ (self):
        return '<rotating virtual host file logger>'

    def write (self, data, host = ''):
        # Rotate befor data gets writen to file
        self.maybe_rotate(host)
        self.get_file(host).write (data)
        virtual_host_file_logger.maybe_flush(self, host)

    def writeline (self, line, host = ''):
        self.maybe_rotate(host)
        self.get_file(host).writeline (line)
        virtual_host_file_logger.maybe_flush(self, host)

    def writelines (self, lines, host = ''):
        self.maybe_rotate(host)
        self.get_file(host).writelines (lines)
        virtual_host_file_logger.maybe_flush(self, host)

    def maybe_rotate (self, host=''):
        if self.freq and time.time() > self.rotate_when:
            self.rotate()
            self.rotate_when = self.next_backup(self.freq)
        elif self.rotate_when == None and time.time() > self.check_for_expired:
            # Close all files
            try:
                for host in self.files.keys():
                    self.files[host].close()
                    del self.files[host]
            except:
                pass
            self.check_for_expired = self.next_backup('daily')

    def rotate (self):
        # rotate all files
        (yr, mo, day, hr, min, sec, wd, jday, dst) = time.localtime(time.time())
        try:
            for host in self.files.keys():
                oldname = self.files[host].name
                date = '%04d-%02d-%02d' % (yr, mo, day-1)
                newname = self.get_filename(host, date)
                try:
                    open(newname, 'r').close()      # check if file exists
                    # if we get this far then the file exists - we want a unique one
                    newname = self.get_filename(host,
                                      '%s_%02d.%02d.%02d' % (date, hr, min, sec))
                except:
                    pass
                self.files[host].close()
                del self.files[host]
                os.rename(oldname, newname)
        except:
            pass

    # Choose the right open file or open it now
    def get_file (self, host = ''):
        host = self.clean_up_host(host)
        
        # Return the file (open if necessary)
        if not self.files.has_key(host):
            filename = self.get_filename(host)
            dir = os.path.dirname(filename)
            # if dir doesn't exist, make it
            if not os.path.isdir(dir):
                os.makedirs(dir)
            file = open(filename, self.mode)
            self.files[host] = file
        else:
            file = self.files[host]
        return file


# SuperLogger will only work if you have installed the SuperLogger product
# into Zope.
import os.path
import ZODB, ZODB.FileStorage

def SuperLogger_variables(zope_app = None, db_file = 'data.fs'):
    dbclose = 0
    # zope_app should be there, but here's some code I was using before I
    #  realised that the ZODB would be open already
    if zope_app is None:
        file = os.path.abspath(db_file)
        db = ZODB.DB(ZODB.FileStorage.FileStorage(file))
        zope_app = db.open().root()['Application']
        dbclose = 1
    else:
        zope_app = zope_app()
    variables = {}
    try:
        sl = zope_app.Control_Panel.SuperLogger
        if hasattr(sl, 'get_variables'):
            variables = sl.get_variables()
        else:
            variables = {}
    finally:
        if dbclose:
            db.close()
    return variables

def SuperLogger(file, zope_module):
    v = SuperLogger_variables(getattr(zope_module, 'app', None))
    virtualhost = v.get('virtualhost', 0)
    rotational = v.get('rotational', 0)
    frequency = v.get('frequency', 0)
    if virtualhost:
        fileformat = v.get('fileformat', 0)
        whitelist = v.get('whitelist', 0)
        if whitelist:
            hosts = v.get('hosts', [])
        else:
            hosts = []
        if rotational:
            return rotating_virtual_host_file_logger(file, frequency, hosts=hosts, file_format=fileformat)
        else:
            return virtual_host_file_logger(file, hosts=hosts, file_format=fileformat)
    else:
        if rotational:
            return rotating_file_logger(file, frequency)
        else:
            return file_logger(file)



# syslog is a line-oriented log protocol - this class would be
# appropriate for FTP or HTTP logs, but not for dumping stderr to.
            
# TODO: a simple safety wrapper that will ensure that the line sent
# to syslog is reasonable.
            
# TODO: async version of syslog_client: now, log entries use blocking
# send()

import m_syslog
syslog_logger = m_syslog.syslog_client

class syslog_logger (m_syslog.syslog_client):

    svc_name = 'medusa'
    pid_str  = str(os.getpid())

    def __init__ (self, address, facility='user'):
        m_syslog.syslog_client.__init__ (self, address)
        self.facility = m_syslog.facility_names[facility]
        self.address=address
        
    def __repr__ (self):
        return '<syslog logger address=%s>' % (repr(self.address))
        
    def log (self, message):
        m_syslog.syslog_client.log (
            self,
            '%s[%s]: %s' % (self.svc_name, self.pid_str, message),
            facility=self.facility,
            priority=m_syslog.LOG_INFO
            )
        
        # log to a stream socket, asynchronously
        
class socket_logger (asynchat.async_chat):

    def __init__ (self, address):
    
        if type(address) == type(''):
            self.create_socket (socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
            
        self.connect (address)
        self.address = address
        
    def __repr__ (self):
        return '<socket logger: address=%s>' % (self.address)
        
    def log (self, message):
        if message[-2:] != '\r\n':
            self.socket.push (message + '\r\n')
        else:
            self.socket.push (message)
            
            # log to multiple places
class multi_logger:
    def __init__ (self, loggers):
        self.loggers = loggers
        
    def __repr__ (self):
        return '<multi logger: %s>' % (repr(self.loggers))
        
    def log (self, message):
        for logger in self.loggers:
            logger.log (message)
            
class resolving_logger:
    """Feed (ip, message) combinations into this logger to get a
    resolved hostname in front of the message.  The message will not
    be logged until the PTR request finishes (or fails)."""
    
    def __init__ (self, resolver, logger, host):
        self.resolver = resolver
        self.logger = logger
        self.host = host
        
    class logger_thunk:
        def __init__ (self, message, logger):
            self.message = message
            self.logger = logger
            
        def __call__ (self, host, ttl, answer):
            if not answer:
                answer = host
            self.logger.log ('%s%s' % (answer, self.message), self.host)
            
    def log (self, ip, message, host = ''):
        self.resolver.resolve_ptr (
                ip,
                self.logger_thunk (
                        message,
                        self.logger,
                        host
                        )
                )
        
class unresolving_logger:
    "Just in case you don't want to resolve"
    def __init__ (self, logger):
        self.logger = logger
        
    def log (self, ip, message, host = ''):
        self.logger.log ('%s%s' % (ip, message), host)
        
        
def strip_eol (line):
    while line and line[-1] in '\r\n':
        line = line[:-1]
    return line
    
class tail_logger:
    "Keep track of the last <size> log messages"
    def __init__ (self, logger, size=500):
        self.size = size
        self.logger = logger
        self.messages = []
        
    def log (self, message):
        self.messages.append (strip_eol (message))
        if len (self.messages) > self.size:
            del self.messages[0]
        self.logger.log (message)
